#r "System.Windows.Forms"
using System.Windows.Forms;
using System.Linq;
using System.Resources;
using System.IO;

string resxPath = Path.Combine(Path.GetDirectoryName(ScriptFilePath), Application.Settings.LanguageCode == "ja" ? "Resources.ja.resx" : "Resources.resx");
ResXResourceSet resx = new ResXResourceSet(resxPath);

/// <summary>
/// 部分図にグループを追加
/// </summary>
/// <param name="drawing">部分図</param>
/// <param name="shape">グループ内の図形</param>
/// <param name="isEmptyLayer">レイヤなしか</param>
/// <returns>グループ図形</returns>
/// <remarks>
/// 部分図にグループを追加します。
/// </remarks> 
DrawGroupShape CreateDrawGroup(Drawing drawing, Shape shape, bool isEmptyLayer)
{
    // レイヤ状態データ
    var layerStateMap = new Dictionary<Layer, ShapeState>();

    // 図形のレイヤが編集可能以外の場合
    if(shape.LayerID != 0 && shape.Layer.State != ShapeState.Editable)
    {
        // 前の状態を保持しておく
        layerStateMap.Add(shape.Layer, shape.Layer.State);

        // 一時的に編集可能に変更する
        shape.Layer.State = ShapeState.Editable;
    }
    // レイヤなしの場合
    else if(shape.LayerID == 0)
    {
        SaveGroupLayerState(shape, layerStateMap);
    }

    Shape[] shapes = new Shape[] { shape };
    DrawGroupShape drawGroup = drawing.Shapes.AddDrawGroup(shapes, isEmptyLayer);

    // 変更したレイヤ状態を元に戻す
    LoadLayerState(layerStateMap);

    return drawGroup;
}

/// <summary>
/// グループ内図形のレイヤ状態保存
/// </summary>
/// <param name="shape">図形</param>
/// <param name="layerStateMap">レイヤ状態</param>
/// <returns></returns>
/// <remarks>
/// グループ内の図形のレイヤ状態を保存して、レイヤを編集可能に変更します。
/// </remarks> 
void SaveGroupLayerState(Shape shape, Dictionary<Layer, ShapeState>  layerStateMap)
{
    Shape[] explodedShapes = shape.GetExplodedShapes();

    // グループ内のレイヤ状態を保持して編集可能に変更する
    foreach(Shape exShape in explodedShapes)
    {
        Layer layer = ActiveDocument.LayerTable.GetLayerByID(exShape.LayerID);
        if(layer.ID != 0 && layer.State != ShapeState.Editable)
        {
            if(!layerStateMap.ContainsKey(layer))
            {
                layerStateMap.Add(layer, layer.State);
                layer.State = ShapeState.Editable;
            }
        }
    }
}

/// <summary>
/// レイヤ状態のロード
/// </summary>
/// <param name="layerStateMap">レイヤ状態</param>
/// <returns></returns>
/// <remarks>
/// レイヤ状態を戻します。
/// </remarks> 
void LoadLayerState(Dictionary<Layer, ShapeState> layerStateMap)
{
    // レイヤ状態を元に戻す
    foreach(KeyValuePair<Layer, ShapeState> item in layerStateMap)
        item.Key.State = item.Value;
}

/// <summary>
/// 図形をグループに追加
/// </summary>
/// <param name="drawGroup">グループ</param>
/// <param name="shape">図形</param>
/// <returns></returns>
/// <remarks>
/// グループに図形を追加します。
/// </remarks> 
void AddShapeToGroup(DrawGroupShape drawGroup, Shape shape)
{
    // レイヤ状態データ
    var layerStateMap = new Dictionary<Layer, ShapeState>();

    // グループのレイヤが編集可能以外の場合
    if(drawGroup.LayerID != 0 && drawGroup.Layer.State != ShapeState.Editable)
    {
        // 前の状態を保持しておく
        layerStateMap.Add(drawGroup.Layer, drawGroup.Layer.State);

        // 一時的に編集可能に変更する
        drawGroup.Layer.State = ShapeState.Editable;
    }
    // レイヤなしの場合
    else if(drawGroup.LayerID == 0)
    {
        SaveGroupLayerState(drawGroup, layerStateMap);
    }

    // 図形のレイヤが編集可能以外の場合
    if(shape.LayerID != 0 && drawGroup.LayerID != shape.LayerID && shape.Layer.State != ShapeState.Editable)
    {
        // 前の状態を保持しておく
        layerStateMap.Add(shape.Layer, shape.Layer.State);

        // 一時的に編集可能に変更する
        shape.Layer.State = ShapeState.Editable;
    }
    // レイヤなしの場合
    else if(shape.LayerID == 0)
    {
        SaveGroupLayerState(shape, layerStateMap);
    }

    drawGroup.Shapes.Add(shape);

    // 変更したレイヤ状態を元に戻す
    LoadLayerState(layerStateMap);
}

/// <summary>
/// 子グループ作成処理
/// </summary>
/// <param name="ensureGroup">元のグループ</param>
/// <param name="shapeGroupMap">コピー元、コピー先図形グループマッピング</param>
/// <param name="newDrawing">新しい部分図</param>
/// <returns></returns>
/// <remarks>
/// 子グループを作成します。孫グループが作成されていない場合は再帰的に処理します。
/// </remarks> 
void EnsureChildGroup(DrawGroupShape ensureGroup, Dictionary<DrawGroupShape, DrawGroupShape>  shapeGroupMap, Drawing newDrawing)
{
    // 作成済みの場合
    if(shapeGroupMap.ContainsKey(ensureGroup))
        return;

    // グループ内の図形を列挙
    foreach(Shape shape in ensureGroup.Shapes)
    {
        if(!(shape is DrawGroupShape))
            continue;

        // 孫グループが作成されていない
        if(!(shapeGroupMap.ContainsKey((DrawGroupShape)shape)))
            // 孫グループ作成
            EnsureChildGroup((DrawGroupShape)shape, shapeGroupMap, newDrawing);

        bool emptyLayer = ensureGroup.LayerID == 0 ? true : false;
        // グループを追加
        var newGroup = CreateDrawGroup(newDrawing, shapeGroupMap[(DrawGroupShape)shape], emptyLayer);

        // レイヤを設定
        if(!emptyLayer)
            newGroup.LayerID = ensureGroup.LayerID;

        // 属性をコピー
        foreach(ShapeAttribute att in ensureGroup.Attributes)
            newGroup.Attributes.Add(att.Name, att.Value);

        shapeGroupMap.Add(ensureGroup, newGroup);
        return;
    }
    return;
}

var node = Application.DrawingBar.SelectedNode;
if(node == null || node.Drawing == null)
{
    MessageBox.Show(resx.GetString("ConfirmMessage1"));
    return;
}

// 部分図レイアウトは複製しない
if(node.Drawing is Part && node.Layout != null)
{
    MessageBox.Show(resx.GetString("ConfirmMessage2"));
    return;
}

// 部分図ツリーウィンドウで選択されている部分図を取得します
var selectedDrawing = node.Drawing;

// 追加する部分図の名前は、元の部分図名の後ろに (2) を付けます
int n = 2;
while(ActiveDocument.Drawings.Any(drawing => drawing.Name == n.ToString(selectedDrawing.Name + " (0)")))
    n++;

// UNDO 処理を開始します
Application.ActiveDocument.UndoManager.BeginUndoUnit();

Drawing newDrawing = null;

// 部分図の場合
if(selectedDrawing is Part)
{
    // 新しい部分図を追加します
    var newPart = ActiveDocument.Drawings.AddPart(n.ToString(selectedDrawing.Name + " (0)"));
    newPart.Printable = selectedDrawing.Printable;
    newPart.Description = selectedDrawing.Description;

    newDrawing = newPart;
}
// 用紙の場合
else if(selectedDrawing is Paper)
{
    // 新しい用紙を追加します
    var newPaper = ActiveDocument.Drawings.AddPaper(n.ToString(selectedDrawing.Name + " (0)"));

    newPaper.PaperType = ((Paper)selectedDrawing).PaperType;
    newPaper.Landscape = ((Paper)selectedDrawing).Landscape;

    if(((Paper)selectedDrawing).PaperType == PaperType.Custom)
    {
        newPaper.PaperWidth = ((Paper)selectedDrawing).PaperWidth;
        newPaper.PaperHeight = ((Paper)selectedDrawing).PaperHeight;
    }
    newPaper.Printable = ((Paper)selectedDrawing).Printable;
    newPaper.PrintScale = ((Paper)selectedDrawing).PrintScale;
    newPaper.Description = ((Paper)selectedDrawing).Description;

    newDrawing = newPaper;
}

// コピー元、コピー先図形グループマッピング Key:src Value:dst
var shapeGroupMap = new Dictionary<DrawGroupShape, DrawGroupShape>();

// 子グループ、親グループマッピング Key:child Value:parent
var groupInGroupMap = new Dictionary<DrawGroupShape, DrawGroupShape>();

// 新しい部分図に図形をコピーします
foreach(Shape shape in selectedDrawing.Shapes)
{
    // 作図グループの場合
    if(shape is DrawGroupShape)
    {
        // 親グループがある場合は親子関係を保持しておく
        if(shape.DrawGroup != null)
            groupInGroupMap.Add((DrawGroupShape)shape, shape.DrawGroup);
    }
    // OLE オブジェクト以外
    else if(!(shape is OleObjectShape))
    {
        // 図形を新しい部分図に追加します
        var newShape = newDrawing.Shapes.Add(shape.Clone());

        // グループ内の図形の場合
        if(shape.DrawGroup != null)
        {
            // グループ作成済み
            if(shapeGroupMap.ContainsKey(shape.DrawGroup))
            {
                AddShapeToGroup(shapeGroupMap[shape.DrawGroup], newShape);
            }
            // グループ未作成
            else
            {
                bool emptyLayer = shape.DrawGroup.LayerID == 0 ? true : false;
                // グループを追加
                var newGroup = CreateDrawGroup(newDrawing, newShape, emptyLayer);

                // レイヤの設定
                if(!emptyLayer)
                    newGroup.LayerID = shape.DrawGroup.LayerID;

                // 属性をコピー
                foreach(ShapeAttribute att in shape.DrawGroup.Attributes)
                    newGroup.Attributes.Add(att.Name, att.Value);

                // グループのマッピングに追加
                shapeGroupMap.Add(shape.DrawGroup, newGroup);
            }
        }
    }
}

// グループ内グループの親子関係を作成します
foreach(KeyValuePair<DrawGroupShape, DrawGroupShape> item in groupInGroupMap)
{
    var childGroup = item.Key;
    var parentGroup = item.Value;

    // 子グループが既に作成済みの場合
    if(shapeGroupMap.ContainsKey(childGroup))
    {
        // 親グループも作成済みの場合
        if(shapeGroupMap.ContainsKey(parentGroup))
        {
            // 親グループに子グループを追加
            if(shapeGroupMap[parentGroup].Shapes.IndexOf(shapeGroupMap[childGroup]) == -1)
                AddShapeToGroup(shapeGroupMap[parentGroup], shapeGroupMap[childGroup]);
        }
        // 親グループ未作成
        else
        {
            bool emptyLayer = parentGroup.LayerID == 0 ? true : false;
            // グループを追加
            var newGroup = CreateDrawGroup(newDrawing, shapeGroupMap[childGroup], emptyLayer);

            // レイヤを設定
            if(!emptyLayer)
                newGroup.LayerID = parentGroup.LayerID;

            // 属性をコピー
            foreach(ShapeAttribute att in parentGroup.Attributes)
                newGroup.Attributes.Add(att.Name, att.Value);

            shapeGroupMap.Add(parentGroup, newGroup);
        }
    }
    // 子グループ未作成
    else
    {
        EnsureChildGroup(childGroup, shapeGroupMap, newDrawing);
    }
}

// UNDO 処理を終了します。
Application.ActiveDocument.UndoManager.EndUndoUnit();
